@@ -126,6 +126,28 @@ def get_version_data(session, url, authenticated=None):
         except KeyError:
             pass
 
+        # Older Ironic does not actually return a discovery document for the
+        # single version discovery endpoint, which confuses the single-version
+        # fallback logic. While there are no known other services returning
+        # min/max ranges using headers instead of body, this is done in a
+        # non-Ironic specific manner just in case.
+        # The existence of this support should not be an indication to any
+        # OpenStack services that they should ADD this.
+        if 'id' in body_resp:
+            body_resp['status'] = Status.CURRENT
+            for header in resp.headers:
+                # We lose the case-insensitive quality here
+                header = header.lower()
+                if not header.startswith('x-openstack'):
+                    continue
+                # Once the body starts having these values, stop overriding
+                # with the header values
+                if header.endswith('api-minimum-version'):
+                    body_resp.setdefault('min_version', resp.headers[header])
+                if header.endswith('api-maximum-version'):
+                    body_resp.setdefault('version', resp.headers[header])
+            return [body_resp]
+
     err_text = resp.text[:50] + '...' if len(resp.text) > 50 else resp.text
     raise exceptions.DiscoveryFailure('Invalid Response - Bad version data '
                                       'returned: %s' % err_text)
@@ -323,7 +345,7 @@ def version_to_string(version):
     return ".".join(map(_str_or_latest, version))
 
 
-def _version_between(min_version, max_version, candidate):
+def version_between(min_version, max_version, candidate):
     """Determine whether a candidate version is within a specified range.
 
     :param min_version: The minimum version that is acceptable.
@@ -417,10 +439,28 @@ def _combine_relative_url(discovery_url, version_url):
     parsed_version_url = urllib.parse.urlparse(url)
     parsed_discovery_url = urllib.parse.urlparse(discovery_url)
 
+    # The services can override the version_url with some config options.(for
+    # example, In Keystone, Cinder and Glance, the option is "public_endpoint",
+    # and "compute_link_prefix", "network_link_prefix" in Nova and Neutron.
+    # In this case, it's hard to distinguish which part in version_url is
+    # useful for discovery_url , so here we just get the version from
+    # version_url and then add it into the discovery_url if needed.
+    path = parsed_version_url.path
+    if parsed_discovery_url.netloc != parsed_version_url.netloc:
+        version = version_url.rstrip('/').split('/')[-1]
+        url_path = parsed_discovery_url.path.rstrip('/')
+        if not url_path.endswith(version):
+            path = url_path + '/' + version
+            if version_url.endswith('/'):
+                # add '/' back to keep backward compatibility.
+                path = path + '/'
+        else:
+            path = parsed_discovery_url.path
+
     return urllib.parse.ParseResult(
         parsed_discovery_url.scheme,
         parsed_discovery_url.netloc,
-        parsed_version_url.path,
+        path,
         parsed_version_url.params,
         parsed_version_url.query,
         parsed_version_url.fragment).geturl()
@@ -462,7 +502,7 @@ class Status(object):
         """
         status = raw_status.upper()
         if status == 'STABLE':
-            status = 'CURRENT'
+            status = cls.CURRENT
         if status not in cls.KNOWN:
             status = cls.UNKNOWN
         return status
@@ -714,7 +754,7 @@ class Discover(object):
             if _latest_soft_match(min_version, data['version']):
                 return data
             # Only validate version bounds if versions were specified
-            if min_version and max_version and _version_between(
+            if min_version and max_version and version_between(
                     min_version, max_version, data['version']):
                 return data
 
@@ -724,7 +764,7 @@ class Discover(object):
         # so that they know what version they got. We can return the first
         # entry from version_data, because the user hasn't requested anything
         # different.
-        if no_version and url:
+        if no_version and url and len(version_data) > 0:
             return version_data[0]
 
         # We couldn't find a match.
@@ -1103,6 +1143,9 @@ class EndpointData(object):
                     " and {max_version}".format(
                         min_version=version_to_string(min_version),
                         max_version=version_to_string(max_version)))
+            else:
+                raise exceptions.DiscoveryFailure(
+                    "No version data found remotely at all")
 
         self.min_microversion = discovered_data['min_microversion']
         self.max_microversion = discovered_data['max_microversion']
@@ -1235,7 +1278,7 @@ class EndpointData(object):
         else:
             # `is_between` means version bounds were specified *and* the URL
             # version is between them.
-            is_between = min_version and max_version and _version_between(
+            is_between = min_version and max_version and version_between(
                 min_version, max_version, url_version)
             exact_match = (is_between and max_version and
                            max_version[0] == url_version[0])
@@ -1372,6 +1415,18 @@ def get_discovery(session, url, cache=None, authenticated=False):
     if session.auth and hasattr(session.auth, '_discovery_cache'):
         caches.append(session.auth._discovery_cache)
 
+    # https://example.com and https://example.com/ should be treated the same
+    # for caching purposes.
+    parsed_url = urllib.parse.urlparse(url)
+    if parsed_url.path in ('', '/'):
+        url = urllib.parse.ParseResult(
+            parsed_url.scheme,
+            parsed_url.netloc,
+            '',
+            parsed_url.params,
+            parsed_url.query,
+            parsed_url.fragment).geturl()
+
     for cache in caches:
         disc = cache.get(url)
 
